/**
 * Orthanc - A Lightweight, RESTful DICOM Store
 * Copyright (C) 2012-2016 Sebastien Jodogne, Medical Physics
 * Department, University Hospital of Liege, Belgium
 * Copyright (C) 2017-2023 Osimis S.A., Belgium
 * Copyright (C) 2021-2023 Sebastien Jodogne, ICTEAM UCLouvain, Belgium
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 **/


#pragma once


namespace OrthancDatabases
{
  enum ValueType
  {
    ValueType_BinaryString,
    ValueType_InputFile,
    ValueType_Integer64,
    ValueType_Null,
    ValueType_ResultFile,
    ValueType_Utf8String
  };

  enum Dialect
  {
    Dialect_MySQL,
    Dialect_PostgreSQL,
    Dialect_SQLite,
    Dialect_MSSQL,
    Dialect_Unknown
  };

  enum TransactionType
  {
    TransactionType_ReadWrite,
    TransactionType_ReadOnly,  // Should only arise with Orthanc SDK >= 1.9.2 in the index plugin
    TransactionType_Implicit   // Should only arise with Orthanc SDK <= 1.9.1
  };

  const char* EnumerationToString(ValueType type);
}
